// File generated with precis-tools version 0.1.9

static VIRAMA: [Codepoints; 33] = [
	Codepoints::Single(0x094d),
	Codepoints::Single(0x09cd),
	Codepoints::Single(0x0a4d),
	Codepoints::Single(0x0acd),
	Codepoints::Single(0x0b4d),
	Codepoints::Single(0x0bcd),
	Codepoints::Single(0x0c4d),
	Codepoints::Single(0x0ccd),
	Codepoints::Single(0x0d4d),
	Codepoints::Single(0x0dca),
	Codepoints::Single(0x0e3a),
	Codepoints::Single(0x0f84),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1039, 0x103a)),
	Codepoints::Single(0x1714),
	Codepoints::Single(0x1734),
	Codepoints::Single(0x17d2),
	Codepoints::Single(0x1a60),
	Codepoints::Single(0x1b44),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1baa, 0x1bab)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1bf2, 0x1bf3)),
	Codepoints::Single(0x2d7f),
	Codepoints::Single(0xa806),
	Codepoints::Single(0xa8c4),
	Codepoints::Single(0xa953),
	Codepoints::Single(0xa9c0),
	Codepoints::Single(0xaaf6),
	Codepoints::Single(0xabed),
	Codepoints::Single(0x10a3f),
	Codepoints::Single(0x11046),
	Codepoints::Single(0x110b9),
	Codepoints::Range(std::ops::RangeInclusive::new(0x11133, 0x11134)),
	Codepoints::Single(0x111c0),
	Codepoints::Single(0x116b6),
];

static GREEK: [Codepoints; 33] = [
	Codepoints::Range(std::ops::RangeInclusive::new(0x0370, 0x0373)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0375, 0x0377)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x037a, 0x037d)),
	Codepoints::Single(0x0384),
	Codepoints::Single(0x0386),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0388, 0x038a)),
	Codepoints::Single(0x038c),
	Codepoints::Range(std::ops::RangeInclusive::new(0x038e, 0x03a1)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x03a3, 0x03e1)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x03f0, 0x03ff)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d26, 0x1d2a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d5d, 0x1d61)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d66, 0x1d6a)),
	Codepoints::Single(0x1dbf),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1f00, 0x1f15)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1f18, 0x1f1d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1f20, 0x1f45)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1f48, 0x1f4d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1f50, 0x1f57)),
	Codepoints::Single(0x1f59),
	Codepoints::Single(0x1f5b),
	Codepoints::Single(0x1f5d),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1f5f, 0x1f7d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1f80, 0x1fb4)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1fb6, 0x1fc4)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1fc6, 0x1fd3)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1fd6, 0x1fdb)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1fdd, 0x1fef)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1ff2, 0x1ff4)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1ff6, 0x1ffe)),
	Codepoints::Single(0x2126),
	Codepoints::Range(std::ops::RangeInclusive::new(0x10140, 0x1018a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d200, 0x1d245)),
];

static HEBREW: [Codepoints; 9] = [
	Codepoints::Range(std::ops::RangeInclusive::new(0x0591, 0x05c7)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x05d0, 0x05ea)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x05f0, 0x05f4)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfb1d, 0xfb36)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfb38, 0xfb3c)),
	Codepoints::Single(0xfb3e),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfb40, 0xfb41)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfb43, 0xfb44)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfb46, 0xfb4f)),
];

static HIRAGANA: [Codepoints; 4] = [
	Codepoints::Range(std::ops::RangeInclusive::new(0x3041, 0x3096)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x309d, 0x309f)),
	Codepoints::Single(0x1b001),
	Codepoints::Single(0x1f200),
];

static KATAKANA: [Codepoints; 8] = [
	Codepoints::Range(std::ops::RangeInclusive::new(0x30a1, 0x30fa)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x30fd, 0x30ff)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x31f0, 0x31ff)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x32d0, 0x32fe)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x3300, 0x3357)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xff66, 0xff6f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xff71, 0xff9d)),
	Codepoints::Single(0x1b000),
];

static HAN: [Codepoints; 15] = [
	Codepoints::Range(std::ops::RangeInclusive::new(0x2e80, 0x2e99)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2e9b, 0x2ef3)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2f00, 0x2fd5)),
	Codepoints::Single(0x3005),
	Codepoints::Single(0x3007),
	Codepoints::Range(std::ops::RangeInclusive::new(0x3021, 0x3029)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x3038, 0x303b)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x3400, 0x4db5)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x4e00, 0x9fcc)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xf900, 0xfa6d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfa70, 0xfad9)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x20000, 0x2a6d6)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2a700, 0x2b734)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2b740, 0x2b81d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2f800, 0x2fa1d)),
];

static DUAL_JOINING: [Codepoints; 41] = [
	Codepoints::Single(0x0620),
	Codepoints::Single(0x0626),
	Codepoints::Single(0x0628),
	Codepoints::Range(std::ops::RangeInclusive::new(0x062a, 0x062e)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0633, 0x063f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0641, 0x0647)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0649, 0x064a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x066e, 0x066f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0678, 0x0687)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x069a, 0x06bf)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06c1, 0x06c2)),
	Codepoints::Single(0x06cc),
	Codepoints::Single(0x06ce),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06d0, 0x06d1)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06fa, 0x06fc)),
	Codepoints::Single(0x06ff),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0712, 0x0714)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x071a, 0x071d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x071f, 0x0727)),
	Codepoints::Single(0x0729),
	Codepoints::Single(0x072b),
	Codepoints::Range(std::ops::RangeInclusive::new(0x072d, 0x072e)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x074e, 0x0758)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x075c, 0x076a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x076d, 0x0770)),
	Codepoints::Single(0x0772),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0775, 0x0777)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x077a, 0x077f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x07ca, 0x07ea)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0841, 0x0845)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0847, 0x0848)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x084a, 0x084e)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0850, 0x0853)),
	Codepoints::Single(0x0855),
	Codepoints::Single(0x08a0),
	Codepoints::Range(std::ops::RangeInclusive::new(0x08a2, 0x08a9)),
	Codepoints::Single(0x1807),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1820, 0x1877)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1887, 0x18a8)),
	Codepoints::Single(0x18aa),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa840, 0xa871)),
];

static LEFT_JOINING: [Codepoints; 1] = [
	Codepoints::Single(0xa872),
];

static RIGHT_JOINING: [Codepoints; 34] = [
	Codepoints::Range(std::ops::RangeInclusive::new(0x0622, 0x0625)),
	Codepoints::Single(0x0627),
	Codepoints::Single(0x0629),
	Codepoints::Range(std::ops::RangeInclusive::new(0x062f, 0x0632)),
	Codepoints::Single(0x0648),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0671, 0x0673)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0675, 0x0677)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0688, 0x0699)),
	Codepoints::Single(0x06c0),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06c3, 0x06cb)),
	Codepoints::Single(0x06cd),
	Codepoints::Single(0x06cf),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06d2, 0x06d3)),
	Codepoints::Single(0x06d5),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06ee, 0x06ef)),
	Codepoints::Single(0x0710),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0715, 0x0719)),
	Codepoints::Single(0x071e),
	Codepoints::Single(0x0728),
	Codepoints::Single(0x072a),
	Codepoints::Single(0x072c),
	Codepoints::Single(0x072f),
	Codepoints::Single(0x074d),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0759, 0x075b)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x076b, 0x076c)),
	Codepoints::Single(0x0771),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0773, 0x0774)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0778, 0x0779)),
	Codepoints::Single(0x0840),
	Codepoints::Single(0x0846),
	Codepoints::Single(0x0849),
	Codepoints::Single(0x084f),
	Codepoints::Single(0x0854),
	Codepoints::Range(std::ops::RangeInclusive::new(0x08aa, 0x08ac)),
];

static TRANSPARENT: [Codepoints; 232] = [
	Codepoints::Single(0x00ad),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0300, 0x036f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0483, 0x0489)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0591, 0x05bd)),
	Codepoints::Single(0x05bf),
	Codepoints::Range(std::ops::RangeInclusive::new(0x05c1, 0x05c2)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x05c4, 0x05c5)),
	Codepoints::Single(0x05c7),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0610, 0x061a)),
	Codepoints::Single(0x061c),
	Codepoints::Range(std::ops::RangeInclusive::new(0x064b, 0x065f)),
	Codepoints::Single(0x0670),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06d6, 0x06dc)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06df, 0x06e4)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06e7, 0x06e8)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x06ea, 0x06ed)),
	Codepoints::Single(0x070f),
	Codepoints::Single(0x0711),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0730, 0x074a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x07a6, 0x07b0)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x07eb, 0x07f3)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0816, 0x0819)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x081b, 0x0823)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0825, 0x0827)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0829, 0x082d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0859, 0x085b)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x08e4, 0x08fe)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0900, 0x0902)),
	Codepoints::Single(0x093a),
	Codepoints::Single(0x093c),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0941, 0x0948)),
	Codepoints::Single(0x094d),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0951, 0x0957)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0962, 0x0963)),
	Codepoints::Single(0x0981),
	Codepoints::Single(0x09bc),
	Codepoints::Range(std::ops::RangeInclusive::new(0x09c1, 0x09c4)),
	Codepoints::Single(0x09cd),
	Codepoints::Range(std::ops::RangeInclusive::new(0x09e2, 0x09e3)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0a01, 0x0a02)),
	Codepoints::Single(0x0a3c),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0a41, 0x0a42)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0a47, 0x0a48)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0a4b, 0x0a4d)),
	Codepoints::Single(0x0a51),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0a70, 0x0a71)),
	Codepoints::Single(0x0a75),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0a81, 0x0a82)),
	Codepoints::Single(0x0abc),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0ac1, 0x0ac5)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0ac7, 0x0ac8)),
	Codepoints::Single(0x0acd),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0ae2, 0x0ae3)),
	Codepoints::Single(0x0b01),
	Codepoints::Single(0x0b3c),
	Codepoints::Single(0x0b3f),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0b41, 0x0b44)),
	Codepoints::Single(0x0b4d),
	Codepoints::Single(0x0b56),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0b62, 0x0b63)),
	Codepoints::Single(0x0b82),
	Codepoints::Single(0x0bc0),
	Codepoints::Single(0x0bcd),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0c3e, 0x0c40)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0c46, 0x0c48)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0c4a, 0x0c4d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0c55, 0x0c56)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0c62, 0x0c63)),
	Codepoints::Single(0x0cbc),
	Codepoints::Single(0x0cbf),
	Codepoints::Single(0x0cc6),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0ccc, 0x0ccd)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0ce2, 0x0ce3)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0d41, 0x0d44)),
	Codepoints::Single(0x0d4d),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0d62, 0x0d63)),
	Codepoints::Single(0x0dca),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0dd2, 0x0dd4)),
	Codepoints::Single(0x0dd6),
	Codepoints::Single(0x0e31),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0e34, 0x0e3a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0e47, 0x0e4e)),
	Codepoints::Single(0x0eb1),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0eb4, 0x0eb9)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0ebb, 0x0ebc)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0ec8, 0x0ecd)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0f18, 0x0f19)),
	Codepoints::Single(0x0f35),
	Codepoints::Single(0x0f37),
	Codepoints::Single(0x0f39),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0f71, 0x0f7e)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0f80, 0x0f84)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0f86, 0x0f87)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0f8d, 0x0f97)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x0f99, 0x0fbc)),
	Codepoints::Single(0x0fc6),
	Codepoints::Range(std::ops::RangeInclusive::new(0x102d, 0x1030)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1032, 0x1037)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1039, 0x103a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x103d, 0x103e)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1058, 0x1059)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x105e, 0x1060)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1071, 0x1074)),
	Codepoints::Single(0x1082),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1085, 0x1086)),
	Codepoints::Single(0x108d),
	Codepoints::Single(0x109d),
	Codepoints::Range(std::ops::RangeInclusive::new(0x135d, 0x135f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1712, 0x1714)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1732, 0x1734)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1752, 0x1753)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1772, 0x1773)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x17b4, 0x17b5)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x17b7, 0x17bd)),
	Codepoints::Single(0x17c6),
	Codepoints::Range(std::ops::RangeInclusive::new(0x17c9, 0x17d3)),
	Codepoints::Single(0x17dd),
	Codepoints::Range(std::ops::RangeInclusive::new(0x180b, 0x180d)),
	Codepoints::Single(0x18a9),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1920, 0x1922)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1927, 0x1928)),
	Codepoints::Single(0x1932),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1939, 0x193b)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1a17, 0x1a18)),
	Codepoints::Single(0x1a1b),
	Codepoints::Single(0x1a56),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1a58, 0x1a5e)),
	Codepoints::Single(0x1a60),
	Codepoints::Single(0x1a62),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1a65, 0x1a6c)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1a73, 0x1a7c)),
	Codepoints::Single(0x1a7f),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1b00, 0x1b03)),
	Codepoints::Single(0x1b34),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1b36, 0x1b3a)),
	Codepoints::Single(0x1b3c),
	Codepoints::Single(0x1b42),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1b6b, 0x1b73)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1b80, 0x1b81)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1ba2, 0x1ba5)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1ba8, 0x1ba9)),
	Codepoints::Single(0x1bab),
	Codepoints::Single(0x1be6),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1be8, 0x1be9)),
	Codepoints::Single(0x1bed),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1bef, 0x1bf1)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1c2c, 0x1c33)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1c36, 0x1c37)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1cd0, 0x1cd2)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1cd4, 0x1ce0)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1ce2, 0x1ce8)),
	Codepoints::Single(0x1ced),
	Codepoints::Single(0x1cf4),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1dc0, 0x1de6)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1dfc, 0x1dff)),
	Codepoints::Single(0x200b),
	Codepoints::Range(std::ops::RangeInclusive::new(0x200e, 0x200f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x202a, 0x202e)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2060, 0x2064)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x206a, 0x206f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x20d0, 0x20f0)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2cef, 0x2cf1)),
	Codepoints::Single(0x2d7f),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2de0, 0x2dff)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x302a, 0x302d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x3099, 0x309a)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa66f, 0xa672)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa674, 0xa67d)),
	Codepoints::Single(0xa69f),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa6f0, 0xa6f1)),
	Codepoints::Single(0xa802),
	Codepoints::Single(0xa806),
	Codepoints::Single(0xa80b),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa825, 0xa826)),
	Codepoints::Single(0xa8c4),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa8e0, 0xa8f1)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa926, 0xa92d)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa947, 0xa951)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa980, 0xa982)),
	Codepoints::Single(0xa9b3),
	Codepoints::Range(std::ops::RangeInclusive::new(0xa9b6, 0xa9b9)),
	Codepoints::Single(0xa9bc),
	Codepoints::Range(std::ops::RangeInclusive::new(0xaa29, 0xaa2e)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xaa31, 0xaa32)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xaa35, 0xaa36)),
	Codepoints::Single(0xaa43),
	Codepoints::Single(0xaa4c),
	Codepoints::Single(0xaab0),
	Codepoints::Range(std::ops::RangeInclusive::new(0xaab2, 0xaab4)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xaab7, 0xaab8)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xaabe, 0xaabf)),
	Codepoints::Single(0xaac1),
	Codepoints::Range(std::ops::RangeInclusive::new(0xaaec, 0xaaed)),
	Codepoints::Single(0xaaf6),
	Codepoints::Single(0xabe5),
	Codepoints::Single(0xabe8),
	Codepoints::Single(0xabed),
	Codepoints::Single(0xfb1e),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfe00, 0xfe0f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfe20, 0xfe26)),
	Codepoints::Single(0xfeff),
	Codepoints::Range(std::ops::RangeInclusive::new(0xfff9, 0xfffb)),
	Codepoints::Single(0x101fd),
	Codepoints::Range(std::ops::RangeInclusive::new(0x10a01, 0x10a03)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x10a05, 0x10a06)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x10a0c, 0x10a0f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x10a38, 0x10a3a)),
	Codepoints::Single(0x10a3f),
	Codepoints::Single(0x11001),
	Codepoints::Range(std::ops::RangeInclusive::new(0x11038, 0x11046)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x11080, 0x11081)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x110b3, 0x110b6)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x110b9, 0x110ba)),
	Codepoints::Single(0x110bd),
	Codepoints::Range(std::ops::RangeInclusive::new(0x11100, 0x11102)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x11127, 0x1112b)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1112d, 0x11134)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x11180, 0x11181)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x111b6, 0x111be)),
	Codepoints::Single(0x116ab),
	Codepoints::Single(0x116ad),
	Codepoints::Range(std::ops::RangeInclusive::new(0x116b0, 0x116b5)),
	Codepoints::Single(0x116b7),
	Codepoints::Range(std::ops::RangeInclusive::new(0x16f8f, 0x16f92)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d167, 0x1d169)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d173, 0x1d182)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d185, 0x1d18b)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d1aa, 0x1d1ad)),
	Codepoints::Range(std::ops::RangeInclusive::new(0x1d242, 0x1d244)),
	Codepoints::Single(0xe0001),
	Codepoints::Range(std::ops::RangeInclusive::new(0xe0020, 0xe007f)),
	Codepoints::Range(std::ops::RangeInclusive::new(0xe0100, 0xe01ef)),
];

