// SPDX-FileCopyrightText: 2024 blinry <mail@blinry.org>
// SPDX-FileCopyrightText: 2024 zormit <nt4u@kpvn.de>
// SPDX-FileCopyrightText: 2026 Caleb Maclennan <caleb@alerque.com>
//
// SPDX-License-Identifier: AGPL-3.0-or-later

#![doc = include_str!("../README.md")]

pub mod config;
pub mod daemon;
pub mod document;
pub mod editor;
pub mod editor_connection;
pub mod editor_protocol;
pub mod jsonrpc_forwarder;
pub mod logging;
pub mod ot;
pub mod path;
pub mod peer;
pub mod sandbox;
pub mod types;
pub mod watcher;
pub mod wormhole;
