// File generated with precis-tools version 0.1.9

static SPACE_SEPARATOR: [Codepoints; 7] = [
	Codepoints::Single(0x0020),
	Codepoints::Single(0x00a0),
	Codepoints::Single(0x1680),
	Codepoints::Range(std::ops::RangeInclusive::new(0x2000, 0x200a)),
	Codepoints::Single(0x202f),
	Codepoints::Single(0x205f),
	Codepoints::Single(0x3000),
];

