// File generated with precis-tools version 0.1.9

static WIDE_NARROW_MAPPING: [(Codepoints, u32); 226] = [
	(Codepoints::Single(0x3000), 0x0020),
	(Codepoints::Single(0xff01), 0x0021),
	(Codepoints::Single(0xff02), 0x0022),
	(Codepoints::Single(0xff03), 0x0023),
	(Codepoints::Single(0xff04), 0x0024),
	(Codepoints::Single(0xff05), 0x0025),
	(Codepoints::Single(0xff06), 0x0026),
	(Codepoints::Single(0xff07), 0x0027),
	(Codepoints::Single(0xff08), 0x0028),
	(Codepoints::Single(0xff09), 0x0029),
	(Codepoints::Single(0xff0a), 0x002a),
	(Codepoints::Single(0xff0b), 0x002b),
	(Codepoints::Single(0xff0c), 0x002c),
	(Codepoints::Single(0xff0d), 0x002d),
	(Codepoints::Single(0xff0e), 0x002e),
	(Codepoints::Single(0xff0f), 0x002f),
	(Codepoints::Single(0xff10), 0x0030),
	(Codepoints::Single(0xff11), 0x0031),
	(Codepoints::Single(0xff12), 0x0032),
	(Codepoints::Single(0xff13), 0x0033),
	(Codepoints::Single(0xff14), 0x0034),
	(Codepoints::Single(0xff15), 0x0035),
	(Codepoints::Single(0xff16), 0x0036),
	(Codepoints::Single(0xff17), 0x0037),
	(Codepoints::Single(0xff18), 0x0038),
	(Codepoints::Single(0xff19), 0x0039),
	(Codepoints::Single(0xff1a), 0x003a),
	(Codepoints::Single(0xff1b), 0x003b),
	(Codepoints::Single(0xff1c), 0x003c),
	(Codepoints::Single(0xff1d), 0x003d),
	(Codepoints::Single(0xff1e), 0x003e),
	(Codepoints::Single(0xff1f), 0x003f),
	(Codepoints::Single(0xff20), 0x0040),
	(Codepoints::Single(0xff21), 0x0041),
	(Codepoints::Single(0xff22), 0x0042),
	(Codepoints::Single(0xff23), 0x0043),
	(Codepoints::Single(0xff24), 0x0044),
	(Codepoints::Single(0xff25), 0x0045),
	(Codepoints::Single(0xff26), 0x0046),
	(Codepoints::Single(0xff27), 0x0047),
	(Codepoints::Single(0xff28), 0x0048),
	(Codepoints::Single(0xff29), 0x0049),
	(Codepoints::Single(0xff2a), 0x004a),
	(Codepoints::Single(0xff2b), 0x004b),
	(Codepoints::Single(0xff2c), 0x004c),
	(Codepoints::Single(0xff2d), 0x004d),
	(Codepoints::Single(0xff2e), 0x004e),
	(Codepoints::Single(0xff2f), 0x004f),
	(Codepoints::Single(0xff30), 0x0050),
	(Codepoints::Single(0xff31), 0x0051),
	(Codepoints::Single(0xff32), 0x0052),
	(Codepoints::Single(0xff33), 0x0053),
	(Codepoints::Single(0xff34), 0x0054),
	(Codepoints::Single(0xff35), 0x0055),
	(Codepoints::Single(0xff36), 0x0056),
	(Codepoints::Single(0xff37), 0x0057),
	(Codepoints::Single(0xff38), 0x0058),
	(Codepoints::Single(0xff39), 0x0059),
	(Codepoints::Single(0xff3a), 0x005a),
	(Codepoints::Single(0xff3b), 0x005b),
	(Codepoints::Single(0xff3c), 0x005c),
	(Codepoints::Single(0xff3d), 0x005d),
	(Codepoints::Single(0xff3e), 0x005e),
	(Codepoints::Single(0xff3f), 0x005f),
	(Codepoints::Single(0xff40), 0x0060),
	(Codepoints::Single(0xff41), 0x0061),
	(Codepoints::Single(0xff42), 0x0062),
	(Codepoints::Single(0xff43), 0x0063),
	(Codepoints::Single(0xff44), 0x0064),
	(Codepoints::Single(0xff45), 0x0065),
	(Codepoints::Single(0xff46), 0x0066),
	(Codepoints::Single(0xff47), 0x0067),
	(Codepoints::Single(0xff48), 0x0068),
	(Codepoints::Single(0xff49), 0x0069),
	(Codepoints::Single(0xff4a), 0x006a),
	(Codepoints::Single(0xff4b), 0x006b),
	(Codepoints::Single(0xff4c), 0x006c),
	(Codepoints::Single(0xff4d), 0x006d),
	(Codepoints::Single(0xff4e), 0x006e),
	(Codepoints::Single(0xff4f), 0x006f),
	(Codepoints::Single(0xff50), 0x0070),
	(Codepoints::Single(0xff51), 0x0071),
	(Codepoints::Single(0xff52), 0x0072),
	(Codepoints::Single(0xff53), 0x0073),
	(Codepoints::Single(0xff54), 0x0074),
	(Codepoints::Single(0xff55), 0x0075),
	(Codepoints::Single(0xff56), 0x0076),
	(Codepoints::Single(0xff57), 0x0077),
	(Codepoints::Single(0xff58), 0x0078),
	(Codepoints::Single(0xff59), 0x0079),
	(Codepoints::Single(0xff5a), 0x007a),
	(Codepoints::Single(0xff5b), 0x007b),
	(Codepoints::Single(0xff5c), 0x007c),
	(Codepoints::Single(0xff5d), 0x007d),
	(Codepoints::Single(0xff5e), 0x007e),
	(Codepoints::Single(0xff5f), 0x2985),
	(Codepoints::Single(0xff60), 0x2986),
	(Codepoints::Single(0xff61), 0x3002),
	(Codepoints::Single(0xff62), 0x300c),
	(Codepoints::Single(0xff63), 0x300d),
	(Codepoints::Single(0xff64), 0x3001),
	(Codepoints::Single(0xff65), 0x30fb),
	(Codepoints::Single(0xff66), 0x30f2),
	(Codepoints::Single(0xff67), 0x30a1),
	(Codepoints::Single(0xff68), 0x30a3),
	(Codepoints::Single(0xff69), 0x30a5),
	(Codepoints::Single(0xff6a), 0x30a7),
	(Codepoints::Single(0xff6b), 0x30a9),
	(Codepoints::Single(0xff6c), 0x30e3),
	(Codepoints::Single(0xff6d), 0x30e5),
	(Codepoints::Single(0xff6e), 0x30e7),
	(Codepoints::Single(0xff6f), 0x30c3),
	(Codepoints::Single(0xff70), 0x30fc),
	(Codepoints::Single(0xff71), 0x30a2),
	(Codepoints::Single(0xff72), 0x30a4),
	(Codepoints::Single(0xff73), 0x30a6),
	(Codepoints::Single(0xff74), 0x30a8),
	(Codepoints::Single(0xff75), 0x30aa),
	(Codepoints::Single(0xff76), 0x30ab),
	(Codepoints::Single(0xff77), 0x30ad),
	(Codepoints::Single(0xff78), 0x30af),
	(Codepoints::Single(0xff79), 0x30b1),
	(Codepoints::Single(0xff7a), 0x30b3),
	(Codepoints::Single(0xff7b), 0x30b5),
	(Codepoints::Single(0xff7c), 0x30b7),
	(Codepoints::Single(0xff7d), 0x30b9),
	(Codepoints::Single(0xff7e), 0x30bb),
	(Codepoints::Single(0xff7f), 0x30bd),
	(Codepoints::Single(0xff80), 0x30bf),
	(Codepoints::Single(0xff81), 0x30c1),
	(Codepoints::Single(0xff82), 0x30c4),
	(Codepoints::Single(0xff83), 0x30c6),
	(Codepoints::Single(0xff84), 0x30c8),
	(Codepoints::Single(0xff85), 0x30ca),
	(Codepoints::Single(0xff86), 0x30cb),
	(Codepoints::Single(0xff87), 0x30cc),
	(Codepoints::Single(0xff88), 0x30cd),
	(Codepoints::Single(0xff89), 0x30ce),
	(Codepoints::Single(0xff8a), 0x30cf),
	(Codepoints::Single(0xff8b), 0x30d2),
	(Codepoints::Single(0xff8c), 0x30d5),
	(Codepoints::Single(0xff8d), 0x30d8),
	(Codepoints::Single(0xff8e), 0x30db),
	(Codepoints::Single(0xff8f), 0x30de),
	(Codepoints::Single(0xff90), 0x30df),
	(Codepoints::Single(0xff91), 0x30e0),
	(Codepoints::Single(0xff92), 0x30e1),
	(Codepoints::Single(0xff93), 0x30e2),
	(Codepoints::Single(0xff94), 0x30e4),
	(Codepoints::Single(0xff95), 0x30e6),
	(Codepoints::Single(0xff96), 0x30e8),
	(Codepoints::Single(0xff97), 0x30e9),
	(Codepoints::Single(0xff98), 0x30ea),
	(Codepoints::Single(0xff99), 0x30eb),
	(Codepoints::Single(0xff9a), 0x30ec),
	(Codepoints::Single(0xff9b), 0x30ed),
	(Codepoints::Single(0xff9c), 0x30ef),
	(Codepoints::Single(0xff9d), 0x30f3),
	(Codepoints::Single(0xff9e), 0x3099),
	(Codepoints::Single(0xff9f), 0x309a),
	(Codepoints::Single(0xffa0), 0x3164),
	(Codepoints::Single(0xffa1), 0x3131),
	(Codepoints::Single(0xffa2), 0x3132),
	(Codepoints::Single(0xffa3), 0x3133),
	(Codepoints::Single(0xffa4), 0x3134),
	(Codepoints::Single(0xffa5), 0x3135),
	(Codepoints::Single(0xffa6), 0x3136),
	(Codepoints::Single(0xffa7), 0x3137),
	(Codepoints::Single(0xffa8), 0x3138),
	(Codepoints::Single(0xffa9), 0x3139),
	(Codepoints::Single(0xffaa), 0x313a),
	(Codepoints::Single(0xffab), 0x313b),
	(Codepoints::Single(0xffac), 0x313c),
	(Codepoints::Single(0xffad), 0x313d),
	(Codepoints::Single(0xffae), 0x313e),
	(Codepoints::Single(0xffaf), 0x313f),
	(Codepoints::Single(0xffb0), 0x3140),
	(Codepoints::Single(0xffb1), 0x3141),
	(Codepoints::Single(0xffb2), 0x3142),
	(Codepoints::Single(0xffb3), 0x3143),
	(Codepoints::Single(0xffb4), 0x3144),
	(Codepoints::Single(0xffb5), 0x3145),
	(Codepoints::Single(0xffb6), 0x3146),
	(Codepoints::Single(0xffb7), 0x3147),
	(Codepoints::Single(0xffb8), 0x3148),
	(Codepoints::Single(0xffb9), 0x3149),
	(Codepoints::Single(0xffba), 0x314a),
	(Codepoints::Single(0xffbb), 0x314b),
	(Codepoints::Single(0xffbc), 0x314c),
	(Codepoints::Single(0xffbd), 0x314d),
	(Codepoints::Single(0xffbe), 0x314e),
	(Codepoints::Single(0xffc2), 0x314f),
	(Codepoints::Single(0xffc3), 0x3150),
	(Codepoints::Single(0xffc4), 0x3151),
	(Codepoints::Single(0xffc5), 0x3152),
	(Codepoints::Single(0xffc6), 0x3153),
	(Codepoints::Single(0xffc7), 0x3154),
	(Codepoints::Single(0xffca), 0x3155),
	(Codepoints::Single(0xffcb), 0x3156),
	(Codepoints::Single(0xffcc), 0x3157),
	(Codepoints::Single(0xffcd), 0x3158),
	(Codepoints::Single(0xffce), 0x3159),
	(Codepoints::Single(0xffcf), 0x315a),
	(Codepoints::Single(0xffd2), 0x315b),
	(Codepoints::Single(0xffd3), 0x315c),
	(Codepoints::Single(0xffd4), 0x315d),
	(Codepoints::Single(0xffd5), 0x315e),
	(Codepoints::Single(0xffd6), 0x315f),
	(Codepoints::Single(0xffd7), 0x3160),
	(Codepoints::Single(0xffda), 0x3161),
	(Codepoints::Single(0xffdb), 0x3162),
	(Codepoints::Single(0xffdc), 0x3163),
	(Codepoints::Single(0xffe0), 0x00a2),
	(Codepoints::Single(0xffe1), 0x00a3),
	(Codepoints::Single(0xffe2), 0x00ac),
	(Codepoints::Single(0xffe3), 0x00af),
	(Codepoints::Single(0xffe4), 0x00a6),
	(Codepoints::Single(0xffe5), 0x00a5),
	(Codepoints::Single(0xffe6), 0x20a9),
	(Codepoints::Single(0xffe8), 0x2502),
	(Codepoints::Single(0xffe9), 0x2190),
	(Codepoints::Single(0xffea), 0x2191),
	(Codepoints::Single(0xffeb), 0x2192),
	(Codepoints::Single(0xffec), 0x2193),
	(Codepoints::Single(0xffed), 0x25a0),
	(Codepoints::Single(0xffee), 0x25cb),
];

