//! Networking related utilities

#[cfg_attr(wasm_browser, path = "interfaces/wasm_browser.rs")]
pub mod interfaces;
pub mod ip;
mod ip_family;
pub mod netmon;
#[cfg(not(wasm_browser))]
mod udp;

pub use self::ip_family::IpFamily;
#[cfg(not(wasm_browser))]
pub use self::udp::UdpSocket;
